import { BufferUsage, DataType, Primitive } from "./types";
import { Device as _Device } from "./device";
export declare type ElementArray = number[] | [number, number][] | [number, number, number][] | number[][];
/**
 * Possible data types of element buffers.
 */
export declare type ElementBufferType = DataType.UNSIGNED_BYTE | DataType.UNSIGNED_SHORT | DataType.UNSIGNED_INT;
export declare type ElementBufferTypedArray = Uint8Array | Uint8ClampedArray | Uint16Array | Uint32Array;
export interface ElementBufferTypeToTypedArray {
    [DataType.UNSIGNED_BYTE]: Uint8Array | Uint8ClampedArray;
    [DataType.UNSIGNED_SHORT]: Uint16Array;
    [DataType.UNSIGNED_INT]: Uint32Array;
    [p: number]: ElementBufferTypedArray;
}
export interface ElementBufferOptions {
    usage?: BufferUsage;
}
export interface ElementBufferStoreOptions {
    offset?: number;
}
/**
 * Element buffers contain indices for accessing vertex buffer data. They are,
 * together with vertex buffers part of VertexArray objects.
 */
export declare class ElementBuffer<T extends ElementBufferType> {
    /**
     * Create a new element buffer with given type, primitive, and size.
     */
    static create<T extends ElementBufferType>(dev: _Device, type: T, primitive: Primitive, size: number, {usage}?: ElementBufferOptions): ElementBuffer<T>;
    /**
     * Create a new element buffer from potentially nested array. Infers
     * Primitive from the array's shape:
     *   number[] -> POINTS
     *   [number, number][] -> LINES
     *   [number, number, number][] -> TRIANGLES
     * Array is referenced only for the duration of this call.
     */
    static withArray(dev: _Device, data: ElementArray, options?: ElementBufferOptions): ElementBuffer<DataType.UNSIGNED_INT>;
    /**
     * Create a new element buffer of given type with provided data. Data is
     * referenced only for the duration of this call.
     */
    static withTypedArray<T extends ElementBufferType>(dev: _Device, type: T, primitive: Primitive, data: ElementBufferTypeToTypedArray[T] | number[], {usage}?: ElementBufferOptions): ElementBuffer<T>;
    readonly type: T;
    readonly length: number;
    readonly byteLength: number;
    readonly primitive: Primitive;
    readonly usage: BufferUsage;
    readonly glBuffer: WebGLBuffer | null;
    private gl;
    private constructor();
    /**
     * Reinitialize invalid buffer, eg. after context is lost.
     */
    restore(): void;
    /**
     * Upload new data to buffer. Data is referenced only for the duration of
     * this call.
     */
    store(data: ElementBufferTypeToTypedArray[T] | number[], {offset}?: ElementBufferStoreOptions): this;
    private init();
}
